# pylint: disable=missing-module-docstring
from typing import Any
from pathlib import Path
import yaml
import connexion  # type: ignore
from genelastic.common import ElasticQueryConn


def load_yaml(file_path: Path) -> Any:
    """Load a YAML file and return its content."""
    content = None
    with open(file_path, encoding='utf-8') as f:
        try:
            content = yaml.safe_load(f)
        except yaml.YAMLError as exc:
            raise SystemExit(exc) from exc
    return content


def aggregate_openapi_specs(main_spec_file: Path, additional_spec_path: Path) -> Any:
    """Aggregate OpenAPI specifications from a main file and a directory
    of additional specifications."""
    main_spec = load_yaml(main_spec_file)
    try:
        entries = additional_spec_path.iterdir()
    except OSError as exc:
        raise SystemExit(exc) from exc

    if not 'paths' in main_spec:
        main_spec['paths'] = []

    for entry in entries:
        if not entry.is_file():
            continue

        if not entry.suffix in [".yml", ".yaml"]:
            continue

        content = load_yaml(entry)

        if 'paths' in content:
            main_spec['paths'].update(content['paths'])

    return main_spec


# Initialiser l'application Connexion
connexion_app = connexion.FlaskApp(__name__)
connexion_app.app.config.from_object('src.genelastic.api.settings.Config')

# Initialiser le client Elasticsearch
es_url = connexion_app.app.config['GENAPI_ES_URL']
es_cert_fp = connexion_app.app.config['GENAPI_ES_CERT_FP']
es_api_key = connexion_app.app.config['GENAPI_ES_ENCODED_API_KEY']

connexion_app.app.elastic_query_conn = ElasticQueryConn(es_url, es_cert_fp, api_key=es_api_key)

connexion_app.app.logger.debug("Successfully connected to Elasticsearch server: %s",
                               connexion_app.app.elastic_query_conn.client.info())

# Chemins des fichiers YAML
main_yaml_file = Path(__file__).parents[0] / 'specification.yml'
additional_yaml_dir = Path(__file__).parents[0] / 'extends'

# Charger et combiner les fichiers YAML
yaml_spec = aggregate_openapi_specs(main_yaml_file, additional_yaml_dir)

# Ajouter la spécification vers OpenAPI
connexion_app.add_api(yaml_spec)

if __name__ == '__main__':
    connexion_app.run(debug=True)
